#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>
#include <AxisJoystick.h>
#include <Button.h>
#include <Joystick.h>
#include <SPI.h>
#include <Vector.h>

// serial constants
#define SERIAL_SPEED 9600

// screen constants
#define SCREEN_MOSI   9
#define SCREEN_CLK   10
#define SCREEN_DC    11
#define SCREEN_CS    12
#define SCREEN_RESET 13
#define SCREEN_WIDTH 128
#define SCREEN_HEIGHT 64
#define FRAME_DELAY 50
#define YELLOW_SCREEN_LIMIT 16
#define HIGH_SCORE_X 20
#define NEW_GAME_X 40
#define GENERAL_TEXT_X 30
#define GENERAL_TEXT_Y_ADD 10
#define TEXT_SIZE 1

// joystick constants
#define JOYSTICK_BUTTON_PIN 5
#define JOYSTICK_VRX_PIN A1
#define JOYSTICK_VRY_PIN A2
#define ARDUINO_ADC_MIN 0
#define ARDUINO_ADC_MAX 1023
#define ARDUINO_ADC_CORECTION 512
#define MINIMUM_ABS_DISTANCE 50

// button constants
#define SHOOT_BUTTON_PIN 3

// spaceShip contants
#define SPACE_SHIP_INITIAL_WIDTH 12
#define SPACE_SHIP_INITIAL_HEIGHT 8
#define SPACE_SHIP_INITIAL_LIVES_NO 3
#define SPACE_SHIP_MAX_LIVES_NO 4
#define SPACE_SHIP_LOST_LIFE_REDIMENTION 2
#define SPACE_SHIP_DEFAULT_SPEED 0.01

// alienShip constants
#define ALIENS_SHIPS_ROWS 2
#define ALIENS_SHIPS_FIXED_COLUMNS 3
#define ALIENS_SHIPS_COLUMNS_DEVIATION 2
#define ALIENS_SHIPS_FIXED_ROWS_DISTANCE 10
#define ALIEN_SHIP_DEFAULT_SPEED_X 1
#define ALIEN_SHIP_INITIAL_Y 0
#define ALIEN_SHIP_INITIAL_WIDTH 6
#define ALIEN_SHIP_INITIAL_HEIGHT 4
#define ALIEN_SHIP_FIXED_X_DEVIATION 8
#define ALIEN_SHIP_RANDOM_Y_DEVIATION 2
#define ALIEN_SHIP_RANDOM_X_DEVIATION 4
#define ALIEN_SHIP_RANDOM_WIDTH 8
#define ALIEN_SHIP_RANDOM_HEIGHT 0
#define ALIEN_SHIP_INITIAL_LIVES_NO 1
#define ALIEN_SHIP_LIVES_NO_ADJUSTMENT 32
#define ALIEN_SHIP_LOST_LIFE_REDIMENTION 1
#define ALIEN_SHIP_SHOOT_CHANCE 10
#define ALIENS_SHIPS_MAX_NO 10
#define ALIEN_SHIP_X_ADD 5
#define ALIEN_SHIP_LEFT_DIR_CHANCE 2

// cargoShip constants
#define CARGO_SHIP_WIDTH 16
#define CARGO_SHIP_HEIGHT 4
#define CARGO_SHIP_SPEED 4
#define CARGO_SHIP_INITIAL_WIDTH_ADD 64
#define CARGO_SHIP_RESET_POINT_X -200

// shield constants
#define SHIELDS_NO 3
#define SHIELD_INITIAL_HEIGHT 2
#define SHIELD_INITIAL_WIDTH 16
#define SHIELDS_Y 20
#define SHIELD_HIT_REDIMENTION 4

// projectile constants
#define SPACE_SHIP_PROJECTILE_WIDTH 2
#define SPACE_SHIP_PROJECTILE_HEIGHT 4
#define SPACE_SHIP_PROJECTILE_SPEED 5
#define ALIEN_SHIP_PROJECTILE_WIDTH 2
#define ALIEN_SHIP_PROJECTILE_HEIGHT 2
#define ALIEN_SHIP_PROJECTILE_SPEED_FIXED 2
#define ALIEN_SHIP_PROJECTILE_SPEED_DEVIETION 4
#define PROJECTILES_MAX_NO 10
#define PROJECTILE_MIN_Y -10
#define PROJECITILE_MAX_Y_ADD 20

// direction constants
#define TOWARDS_LEFT_DIR -1
#define TOWARDS_RIGHT_DIR 1
#define TOWARDS_ALLIENS_SHIPS_DIR -1
#define TOWARDS_SPACE_SHIP_DIR 1
#define CHANGE_DIR -1

// general game constants
#define END_GAME_DELAY 10000
#define INITIAL_SCORE 2000
#define BONUS_CARGO_SHIP_HIT_SCORE 10;

struct SpaceShip {
  int16_t x;
  uint8_t y;
  uint8_t width;
  uint8_t height;
  uint8_t livesNo;
};

struct AlienShip {
  int16_t x;
  uint8_t y;
  uint8_t width;
  uint8_t height;
  uint8_t livesNo;
  int8_t dirX;
};

struct CargoShip {
  int16_t x;
  uint8_t y;
  uint8_t width;
  uint8_t height;
};

struct Shield {
  uint8_t x;
  uint8_t y;
  uint8_t width;
  uint8_t height;
};

struct Projectile {
  int16_t x;
  int8_t y;
  uint8_t width;
  uint8_t height;
  int8_t dirY;
  uint8_t speedY;
};

Adafruit_SSD1306 screen(SCREEN_WIDTH, SCREEN_HEIGHT,
  SCREEN_MOSI, SCREEN_CLK, SCREEN_DC, SCREEN_RESET, SCREEN_CS);

AxisJoystick joystick(JOYSTICK_BUTTON_PIN, JOYSTICK_VRX_PIN, JOYSTICK_VRY_PIN);

Button shootButton(SHOOT_BUTTON_PIN);

SpaceShip spaceShip;

AlienShip aliensShipsStorage[ALIENS_SHIPS_MAX_NO];
Vector<AlienShip> aliensShips;

CargoShip cargoShip;

Shield shieldsStorage[SHIELDS_NO];
Vector<Shield> shields;

Projectile projectilesStorage[PROJECTILES_MAX_NO];
Vector<Projectile> projectiles;

uint16_t highScore;
uint16_t score;

inline void initHardwareComponents();
inline void initVectorsStorages();
inline void newGame();
inline void displayNewGameScreen();
inline void waitForNewGameCommand();
inline void initSpaceShip();
inline void initAliensShips();
inline void initCargoShip();
inline void initShields();
inline void moveSpaceShip();
inline void moveAliensShips();
inline void moveCargoShip();
inline void moveProjectiles();
inline void maybeSpaceShipShoot();
inline void maybeAliensShipsShoot();
inline void handleProjectilesSpaceShipCollision();
inline void handleProjectilesAliensShipsCollisions();
inline void handleProjectilesCargoShipCollisions();
inline void handleProjectilesShieldsCollisions();
inline void drawSpaceShip();
inline void drawAliensShips();
inline void drawCargoShip();
inline void drawShields();
inline void drawProjectiles();
inline void gameOver(bool won);
inline void displayGameOverScreen(bool won);
inline bool collisionRectRect(int16_t x1, int16_t y1,
  uint8_t width1, uint8_t height1,
  int16_t x2, int16_t y2,
  uint8_t width2, uint8_t height2,
  bool strict);

void setup() {  
  initHardwareComponents();
  initVectorsStorages();
  highScore = 0;
  newGame();
}

inline void initHardwareComponents() {
  Serial.begin(SERIAL_SPEED);

  if(!screen.begin(SSD1306_SWITCHCAPVCC)) {
    Serial.println(F("SSD1306 allocation failed"));
    for(;;);
  }

  joystick.calibrate(ARDUINO_ADC_MIN, ARDUINO_ADC_MAX);

  shootButton.begin(); 
}

inline void initVectorsStorages() {
  // each vector needs a preallocated memory zone to store its elements 
  aliensShips.setStorage(aliensShipsStorage);
  shields.setStorage(shieldsStorage);
  projectiles.setStorage(projectilesStorage);
}

inline void newGame() {
  displayNewGameScreen();
  waitForNewGameCommand(); 
  
  score = INITIAL_SCORE;
  
  initSpaceShip();
  initAliensShips();
  initCargoShip();
  initShields();
}

inline void displayNewGameScreen() {
  screen.clearDisplay();
  
  screen.setTextSize(TEXT_SIZE);
  screen.setTextColor(SSD1306_WHITE);
  
  screen.setCursor(HIGH_SCORE_X, 0);
  screen.print("High score: ");
  screen.print(highScore);
  
  screen.setCursor(NEW_GAME_X, screen.height() / 2);
  screen.print("New Game");

  screen.display();
}

inline void waitForNewGameCommand() {
  while(!shootButton.pressed());
}

inline void initSpaceShip() {
  spaceShip.x = screen.width() / 2;
  spaceShip.y = screen.height() - 1;
  spaceShip.width = SPACE_SHIP_INITIAL_WIDTH;
  spaceShip.height = SPACE_SHIP_INITIAL_HEIGHT;
  spaceShip.livesNo = SPACE_SHIP_INITIAL_LIVES_NO;
}

inline void initAliensShips() {
  // delete all previously inserted alien ships
  aliensShips.clear();

  // there are multiple rows on which an alien ship can be
  for (uint8_t i = 0; i < ALIENS_SHIPS_ROWS; ++i) {
    // the no of alien ships / row can be different
    uint8_t columnsNo =
      ALIENS_SHIPS_FIXED_COLUMNS + random(ALIENS_SHIPS_COLUMNS_DEVIATION);
    uint8_t alienDistance = screen.width() / columnsNo;
    
    for (int j = 0; j < columnsNo; ++j) {
      // create a new alien ship
      AlienShip alienShip;

      // 2 alien ships can have different sizes
      uint8_t width =
        ALIEN_SHIP_INITIAL_WIDTH + random(ALIEN_SHIP_RANDOM_WIDTH);
      uint8_t height =
        ALIEN_SHIP_INITIAL_HEIGHT + random(ALIEN_SHIP_RANDOM_HEIGHT);

      // the lives no depends on the size of the alien ship
      uint8_t extraLivesNo = (width * height) / ALIEN_SHIP_LIVES_NO_ADJUSTMENT;

      // the position of the alien ship is random within some limits
      alienShip.x = ALIEN_SHIP_X_ADD + alienDistance * j;
      alienShip.y =
        ALIEN_SHIP_INITIAL_Y + random(ALIEN_SHIP_RANDOM_Y_DEVIATION) +
        i * ALIENS_SHIPS_FIXED_ROWS_DISTANCE;

      alienShip.width = width;
      alienShip.height = height;
      alienShip.livesNo = ALIEN_SHIP_INITIAL_LIVES_NO + extraLivesNo;
      
      // the direction in which the alien ship will move initially is random
      alienShip.dirX = random(ALIEN_SHIP_LEFT_DIR_CHANCE) == 0 ?
        TOWARDS_LEFT_DIR : TOWARDS_RIGHT_DIR;

      aliensShips.push_back(alienShip);
    }
  }
}

inline void initCargoShip() {
  cargoShip.x = screen.width() + CARGO_SHIP_INITIAL_WIDTH_ADD;
  cargoShip.y = ALIENS_SHIPS_ROWS * ALIENS_SHIPS_FIXED_ROWS_DISTANCE;
  cargoShip.width = CARGO_SHIP_WIDTH;
  cargoShip.height = CARGO_SHIP_HEIGHT;
}

inline void initShields() {
  // delete all previously inserted shields
  shields.clear();
  
  uint8_t shieldsDistance = screen.width() / (SHIELDS_NO + 1);
  
  for (uint8_t i = 0; i < SHIELDS_NO; ++i) {
    // create a new shields
    Shield shield;
    
    shield.x = shieldsDistance * (i + 1) - SHIELD_INITIAL_WIDTH / 2;
    shield.y = screen.height() - SHIELDS_Y;
    shield.width = SHIELD_INITIAL_WIDTH;
    shield.height = SHIELD_INITIAL_HEIGHT;

    shields.push_back(shield);
  }
}

void loop() {
  // update objects' positions
  moveSpaceShip();
  moveAliensShips();
  moveCargoShip();
  moveProjectiles();
  
  // new projectiles might be generated
  maybeSpaceShipShoot();
  maybeAliensShipsShoot();
  
  // handle collisions
  handleProjectilesSpaceShipCollision();
  handleProjectilesAliensShipsCollisions();
  handleProjectilesCargoShipCollisions();
  handleProjectilesShieldsCollisions();

  // draw all objects and display them on the screen
  screen.clearDisplay();
  drawSpaceShip();
  drawAliensShips();
  drawCargoShip();
  drawShields();
  drawProjectiles();
  screen.display();

  // update score
  score -= 1;
  if (score <= 0) {
    gameOver(false);
    newGame();
  }
  
  delay(FRAME_DELAY);
}

inline void moveSpaceShip() {
  int distance = joystick.readVRx() - ARDUINO_ADC_CORECTION;
  if (abs(distance) < MINIMUM_ABS_DISTANCE) {
    // corection needed because of hardware inaccuracy
    distance = 0;
  }
  
  spaceShip.x += SPACE_SHIP_DEFAULT_SPEED * distance;
  spaceShip.x = max(spaceShip.x, -spaceShip.width / 2);
  spaceShip.x = min(spaceShip.x, screen.width() - spaceShip.width / 2);
}

inline void moveAliensShips() {
  for (int i = 0; i < aliensShips.size(); ++i) {
    aliensShips[i].x +=  aliensShips[i].dirX * ALIEN_SHIP_DEFAULT_SPEED_X;

    if (aliensShips[i].x <= 0 ||
      screen.width() - ALIEN_SHIP_INITIAL_WIDTH - ALIEN_SHIP_RANDOM_WIDTH <=
      aliensShips[i].x) {
      // when screen limit is reached change direction
      aliensShips[i].dirX *= CHANGE_DIR;
    }
  }
}

inline void moveCargoShip() {
  cargoShip.x -= CARGO_SHIP_SPEED;

  if (cargoShip.x <= CARGO_SHIP_RESET_POINT_X) {
    initCargoShip();
  }
}

inline void moveProjectiles() {
  for (uint8_t i = 0; i < projectiles.size(); ++i) {
    projectiles[i].y += projectiles[i].dirY * projectiles[i].speedY;
    
    if (projectiles[i].y  < PROJECTILE_MIN_Y ||
      screen.height() +  PROJECITILE_MAX_Y_ADD < projectiles[i].y) {
      // if the projectile exists the screen delete it
      projectiles.remove(i);
      --i;
    }
  }
}

inline void maybeSpaceShipShoot() {
  if (shootButton.pressed()) {
    // create a new projectile
    Projectile projectile;
    
    projectile.x = spaceShip.x + spaceShip.width / 2;
    projectile.y = spaceShip.y - spaceShip.height - 1;
    projectile.width = SPACE_SHIP_PROJECTILE_WIDTH;
    projectile.height = SPACE_SHIP_PROJECTILE_HEIGHT;
    projectile.dirY = TOWARDS_ALLIENS_SHIPS_DIR;
    projectile.speedY = SPACE_SHIP_PROJECTILE_SPEED;
    
    projectiles.push_back(projectile);
  }
}

inline void maybeAliensShipsShoot() {
  for (uint8_t i = 0; i < aliensShips.size(); ++i) {
    // the aliens shoot at random moments with a probability
    uint8_t shoot = random(ALIEN_SHIP_SHOOT_CHANCE * aliensShips.size());
    
    if (shoot == 0) {
      // create a new projectile
      Projectile projectile;
    
      projectile.x = aliensShips[i].x + aliensShips[i].width / 2;
      projectile.y = aliensShips[i].y + aliensShips[i].height + 1;
      projectile.width = ALIEN_SHIP_PROJECTILE_WIDTH;
      projectile.height = ALIEN_SHIP_PROJECTILE_HEIGHT;
      projectile.dirY = TOWARDS_SPACE_SHIP_DIR;
      
      // 2 alien ships projectiles can have different speeds
      projectile.speedY = ALIEN_SHIP_PROJECTILE_SPEED_FIXED +
        random(ALIEN_SHIP_PROJECTILE_SPEED_DEVIETION);
      
      projectiles.push_back(projectile);
    }
  }
}

inline void handleProjectilesSpaceShipCollision() {
  for (uint8_t i = 0; i < projectiles.size(); ++i) {
    if (projectiles[i].dirY != TOWARDS_SPACE_SHIP_DIR) {
      continue;
    }

    if (collisionRectRect(projectiles[i].x, projectiles[i].y,
     projectiles[i].width, projectiles[i].height,
     spaceShip.x, spaceShip.y,
     spaceShip.width, spaceShip.height,
     true)) {
     // space ship hit 
     --spaceShip.livesNo;
     spaceShip.width -= SPACE_SHIP_LOST_LIFE_REDIMENTION;
     spaceShip.height -= SPACE_SHIP_LOST_LIFE_REDIMENTION;

     if (spaceShip.livesNo == 0) {
       gameOver(false);
       newGame();
     }
          
     projectiles.remove(i);
     --i;
        
     break;
    }
  }
}

inline void handleProjectilesAliensShipsCollisions() {
  for (uint8_t i = 0; i < projectiles.size(); ++i) {
    if (projectiles[i].dirY != TOWARDS_ALLIENS_SHIPS_DIR) {
      continue;
    }

    for (uint8_t j = 0; j < aliensShips.size(); ++j) {
      if (collisionRectRect(projectiles[i].x, projectiles[i].y,
        projectiles[i].width, projectiles[i].height,
        aliensShips[j].x, aliensShips[j].y,
        aliensShips[j].width, aliensShips[j].height,
        false)) {
        // aline ship hit
        --aliensShips[j].livesNo;
        aliensShips[j].width -= ALIEN_SHIP_LOST_LIFE_REDIMENTION;
        aliensShips[j].height -= ALIEN_SHIP_LOST_LIFE_REDIMENTION;

        if (aliensShips[j].livesNo == 0) {
          aliensShips.remove(j);
        }

        if (aliensShips.size() == 0) {
          gameOver(true);
          newGame();
        }
          
        projectiles.remove(i);
        --i;
        
        break;
      }
    }
  }
}

inline void handleProjectilesCargoShipCollisions() {
  if (cargoShip.x + cargoShip.width <= 0) {
    return;
  }
  
  for (uint8_t i = 0; i < projectiles.size(); ++i) {
    if (projectiles[i].dirY != TOWARDS_ALLIENS_SHIPS_DIR) {
      continue;
    }

    if (collisionRectRect(projectiles[i].x, projectiles[i].y,
      projectiles[i].width, projectiles[i].height,
      cargoShip.x, cargoShip.y,
      cargoShip.width, cargoShip.height,
      false)) {
      // cargo ship hit
      cargoShip.x = -cargoShip.width;
      
      if (spaceShip.livesNo < SPACE_SHIP_INITIAL_LIVES_NO) {
        ++spaceShip.livesNo;
        spaceShip.width += SPACE_SHIP_LOST_LIFE_REDIMENTION;
        spaceShip.height += SPACE_SHIP_LOST_LIFE_REDIMENTION;
      }

      score += BONUS_CARGO_SHIP_HIT_SCORE;
      
      projectiles.remove(i);
      --i;
        
      break;
    }
  }
}

inline void handleProjectilesShieldsCollisions() {
  for (uint8_t i = 0; i < projectiles.size(); ++i) {
    for (uint8_t j = 0; j < shields.size(); ++j) {
      if (collisionRectRect(projectiles[i].x, projectiles[i].y,
        projectiles[i].width, projectiles[i].height,
        shields[j].x, shields[j].y,
        shields[j].width, shields[j].height,
        false)) {
        // shield hit
        shields[j].width -= SHIELD_HIT_REDIMENTION;

        if (shields[j].width <= 0) {
          shields.remove(j);
        }
          
        projectiles.remove(i);
        --i;
        
        break;
      }
    }
  }
}

inline void drawSpaceShip() {
  screen.fillTriangle(
    spaceShip.x, spaceShip.y,
    spaceShip.x + spaceShip.width, spaceShip.y,
    spaceShip.x + spaceShip.width / 2, spaceShip.y - spaceShip.height,
    SSD1306_WHITE);
}

inline void drawAliensShips() {
  for (uint8_t i = 0; i < aliensShips.size(); ++i) {
    screen.drawRect(
      aliensShips[i].x,
      aliensShips[i].y,
      aliensShips[i].width,
      aliensShips[i].height,
      SSD1306_WHITE);

    screen.drawRect(
      aliensShips[i].x + aliensShips[i].width / 4,
      aliensShips[i].y + aliensShips[i].height / 4,
      aliensShips[i].width / 2,
      aliensShips[i].height / 2,
      SSD1306_WHITE);
  }
}

inline void drawCargoShip() {
  screen.drawRect(
    cargoShip.x,
    cargoShip.y,
    cargoShip.width,
    cargoShip.height,
    SSD1306_WHITE);
}

inline void drawShields() {
   for (uint8_t i = 0; i < shields.size(); ++i) {
    screen.fillRect(
      shields[i].x,
      shields[i].y,
      shields[i].width,
      shields[i].height,
      SSD1306_WHITE);
   }
}

inline void drawProjectiles() {
  for (uint8_t i = 0; i < projectiles.size(); ++i) {
    for (uint8_t j = 0; j < projectiles[i].width; ++j) {
      screen.drawLine(
        projectiles[i].x + j,
        projectiles[i].y,
        projectiles[i].x + j,
        projectiles[i].y - projectiles[i].dirY * projectiles[i].height,
        SSD1306_WHITE);
    }
  }
}

inline void gameOver(bool won) {
  displayGameOverScreen(won);
  highScore = won ? max(highScore, score) : highScore;
  delay(END_GAME_DELAY);
}

inline void displayGameOverScreen(bool won) {
  screen.clearDisplay();

  screen.setTextSize(TEXT_SIZE);
  screen.setTextColor(SSD1306_WHITE);

  screen.setCursor(HIGH_SCORE_X, 0);
  screen.print("High score: ");
  screen.print(highScore);
  
  if (won) {
    screen.setCursor(GENERAL_TEXT_X, screen.height() / 2);    
    screen.println("You won :)");
    
    screen.setCursor(GENERAL_TEXT_X, screen.height() / 2 + GENERAL_TEXT_Y_ADD);
    screen.print("Score: ");
    screen.print(score);
  } else {
    screen.setCursor(GENERAL_TEXT_X, screen.height() / 2);
    screen.println("Game over!");
    screen.setCursor(GENERAL_TEXT_X, screen.height() / 2 + GENERAL_TEXT_Y_ADD);
    screen.print("You lost :(");
  }

  screen.display();
}

inline bool collisionRectRect(int16_t x1, int16_t y1,
  uint8_t width1, uint8_t height1,
  int16_t x2, int16_t y2,
  uint8_t width2, uint8_t height2,
  bool strict) {
  bool collision;

  if (strict) {
    collision = x1 < x2 + width2 &&
      x1 + width1 > x2 &&
      y1 < y2 + height2 &&
      y1 + height1 > y2;
  } else {
     collision = x1 <= x2 + width2 &&
      x1 + width1 >= x2 &&
      y1 <= y2 + height2 &&
      y1 + height1 >= y2;
  }

  return collision;
}
